""" - add XPath extractors and assertions for JMeter
 - show warning if no element matched `set-prop` modification
 - do not create hostaliases file when aliases are not specified
 - add `force-parent-sample` option to JMeter executor
 - add `compile-target-java` option for Selenium
 - add dynamic buffer scaling ability to ResultsReader
 - add scheduling ability with `start-at` parameter
 - apply overrides and cli-aliases before creating artifacts-dir
 - fix multiple JMeter warnings when CSV delimiter isn't set
 - add `memory-xmx` option to JMeter to configure JVM heap size

Modules package holds EngineModule implementations

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


